==> superd-0.2.tar.gz: superd-0.2/internal/service/service.go <==
// Copyright 2022 Clayton Craft <clayton@craftyguy.net>
// SPDX-License-Identifier: GPL-3.0-or-later

package service

import (
	"fmt"
	"os"
	"sync"
	"syscall"
	"time"

	goCmd "github.com/go-cmd/cmd"
	"sr.ht/~craftyguy/superd/internal/config"
	"sr.ht/~craftyguy/superd/internal/log"
	"sr.ht/~craftyguy/superd/pkg/api"
)

type Service struct {
	Start   chan bool
	Stop    chan bool
	Details chan api.ServiceDetails
	Config  chan *config.ServiceConfig

	before struct {
		list []string
		sync.RWMutex
	}
	after struct {
		list []string
		sync.RWMutex
	}

	name struct {
		value string
		sync.RWMutex
	}

	autostart struct {
		value bool
		sync.RWMutex
	}

	config *config.ServiceConfig
}

// whether or not the service should be autostarted
func (s *Service) Autostart() bool {
	s.autostart.RLock()
	defer s.autostart.RUnlock()
	return s.autostart.value
}

func (s *Service) Name() string {
	s.name.RLock()
	defer s.name.RUnlock()
	return s.name.value
}

func (s *Service) AddAfter(name string) {
	s.after.Lock()
	defer s.after.Unlock()
	for _, n := range s.after.list {
		if n == name {
			return
		}
	}
	s.after.list = append(s.after.list, name)
}

func (s *Service) AddBefore(name string) {
	s.before.Lock()
	defer s.before.Unlock()
	for _, n := range s.before.list {
		if n == name {
			return
		}
	}
	s.before.list = append(s.before.list, name)
}

func (s *Service) After() []string {
	s.after.RLock()
	defer s.after.RUnlock()
	// TODO: is this really a deep copy?
	return s.after.list
}

func (s *Service) Before() []string {
	s.before.RLock()
	defer s.before.RUnlock()
	return s.before.list
}

var logger *log.Logger

func Init(name string) (s *Service) {
	logger = log.Get()

	s = &Service{
		Start:   make(chan bool),
		Stop:    make(chan bool),
		Details: make(chan api.ServiceDetails),
		Config:  make(chan *config.ServiceConfig),
	}
	s.name.Lock()
	defer s.name.Unlock()
	s.name.value = name

	return
}

var (
	// for mocking in tests
	execCommand = goCmd.NewCmdOptions
	// give 1 second to startup
	// TODO: use systemd.service TimeoutSec, default to 1(?)
	startTimerDuration = 1 * time.Second
)

// startup service routine, e.g. running ExecStartPre, ExecStart. Returns
// api.ServiceFailed on failure, or a serviceProc with references to the
// running process if startup was successful.
func (s *Service) startup() (sProc chan *procDetails, startStatus chan api.ServiceStatus) {
	startStatus = make(chan api.ServiceStatus)
	sProc = make(chan *procDetails)

	go func() {
		defer close(startStatus)
		defer close(sProc)

		// ExecStartPre
		if s.config.ExecStartPreCmd != "" {
			_, ok := s.runCmd(s.config.ExecStartPreCmd, s.config.ExecStartPreCmdArgs)
			if !ok {
				startStatus <- api.ServiceFailed
				return
			}
		}

		// ExecStart
		cmd, ok := s.runCmd(s.config.Cmd, s.config.CmdArgs)
		if !ok {
			startStatus <- api.ServiceFailed
			return
		}

		// ExecStartPost
		if s.config.ExecStartPostCmd != "" {
			_, ok := s.runCmd(s.config.ExecStartPostCmd, s.config.ExecStartPostCmdArgs)
			if !ok {
				startStatus <- api.ServiceFailed
				return
			}
		}

		// send the cmd from ExecStart
		sProc <- cmd
	}()

	return
}

func (s *Service) Run(quit <-chan struct{}) {

	// TODO: should be a manager config option
	// maxRetries := 10

	var (
		stdout         <-chan string
		stderr         <-chan string
		procStatus     <-chan goCmd.Status
		startStatus    <-chan api.ServiceStatus
		onFailComplete chan struct{}
		restartTimer   <-chan time.Time
		details        api.ServiceDetails
		// tries      int

		// used for receiving the started service process details
		procChan chan *procDetails
		p        *procDetails
	)
	details.Status = api.ServiceStopped

	for {
		select {

		// Start (process) requested
		case <-s.Start:

			if details.Status == api.ServiceStarted || details.Status == api.ServiceStarting {
				continue
			}

			procChan, startStatus = s.startup()
			details.Status = api.ServiceStarting

		// new service process info received (e.g. from startup())
		case p = <-procChan:
			startStatus = nil
			procChan = nil
			procStatus = p.procStatus
			logger.Errorf("procStatus: ", fmt.Sprintf("%v", procStatus))
			details.Status = api.ServiceStarted
			details.Pid = p.proc.Status().PID
			stderr = p.stderr
			stdout = p.stdout

		// service startup status
		case ss := <-startStatus:
			startStatus = nil
			procChan = nil
			procStatus = nil
			switch ss {
			case api.ServiceFailed:
				details.Status = api.ServiceFailed
				// Run execOnFailure
				if s.config.ExecOnFailureCmd != "" {
					// Note: a restart is not scheduled until after the
					// ExecOnFailure command has completed
					onFailComplete = make(chan struct{})
					s.execOnFail(onFailComplete)
				} else {
					restartTimer = s.scheduleRestart(details.Status)
				}
			}

		// Stop (process) requested
		case <-s.Stop:
			// set channel to nil to avoid triggering a restart
			procStatus = nil
			details.Status = api.ServiceStopped

			// guard against many rapid calls to stop
			if p == nil {
				continue
			}

			if err := p.proc.Stop(); err != nil {
				logger.Errorln(s.Name(), "stop() returned error: "+err.Error())
				details.Status = api.ServiceFailed
			}
			p = nil

			// Cancel restart and start timer in so that it doesn't trigger and
			// restart the service
			if restartTimer != nil {
				restartTimer = nil
			}

		// Quit (goroutine) requested
		case <-quit:
			// set channel to nil to avoid triggering a restart
			procStatus = nil
			if p != nil {
				if err := p.proc.Stop(); err != nil {
					logger.Errorln(s.Name(), "failed to stop: "+err.Error())
				}
			}
			return

		// Send details/status
		case s.Details <- details:
			continue

		// Status from running process
		case sStatus := <-procStatus:
			// status from go-cmd
			if sStatus.Complete {
				logger.Debugln(s.Name(), "exited with: "+fmt.Sprint(sStatus.Exit))
				if sStatus.Exit != 0 {
					details.Status = api.ServiceFailed

					// fire off any execOnFailure
					if s.config.ExecOnFailureCmd != "" {
						onFailComplete = make(chan struct{})
						s.execOnFail(onFailComplete)

						// don't schedule a restart until after the
						// ExecOnFailure command has completed
						continue
					}
				} else if s.config.Type == config.TypeOneshot {
					// show 'started' for oneshot services
					details.Status = api.ServiceStarted
				} else {
					details.Status = api.ServiceStopped
				}

				// Restart scheduled if necessary
				restartTimer = s.scheduleRestart(details.Status)
			} else {
				logger.Warnln(s.Name(), "received unhandled status: "+fmt.Sprint(s))
			}

		// onFailStatus -- from ExecOnFailure
		case <-onFailComplete:
			onFailComplete = nil
			// Restart scheduled if necessary
			restartTimer = s.scheduleRestart(details.Status)

		// Restart timer
		case <-restartTimer:
			go func() {
				// TODO: ewwwww
				s.Start <- true
			}()

		// Log output from command
		case line, open := <-stderr:
			if !open {
				stderr = nil
				continue
			}
			logger.Debugln(s.Name(), line)
		case line, open := <-stdout:
			if !open {
				stdout = nil
				continue
			}
			logger.Debugln(s.Name(), line)

		// New config received
		case c := <-s.Config:
			s.config = c

			// update details. note that this is only updating fields that
			// aren't changed by starting/stopping the service, or other
			// runtime events related to the supervised service.
			details.Autostart = c.Autostart
			details.ConfigPath = c.Path
			details.Type = string(c.Type)

			// update the autostart field, which is thread safe
			s.autostart.Lock()
			s.autostart.value = c.Autostart
			s.autostart.Unlock()
		}
	}
}

func (s *Service) isUp(pid int) bool {
	up := true

	p, err := os.FindProcess(pid)
	if err != nil {
		up = false
	} else {
		if err := p.Signal(syscall.Signal(0)); err != nil {
			up = false
		}
	}

	return up
}

func (s *Service) oneshotShouldRestart(status api.ServiceStatus) bool {
	// oneshot should only restart if it is set to Always or OnFailure, and
	// only when the status is "failed"
	if status == api.ServiceFailed {
		if s.config.Restart == config.RestartAlways || s.config.Restart == config.RestartFailure {
			return true
		}
	}
	return false
}

func (s *Service) simpleShouldRestart(status api.ServiceStatus) bool {
	if status == api.ServiceFailed {
		if s.config.Restart == config.RestartAlways || s.config.Restart == config.RestartFailure {
			return true
		}
	} else if status == api.ServiceStopped {
		if s.config.Restart == config.RestartAlways {
			return true
		}
	}

	return false
}

// Checks whether a restart should be scheduled based on the given status and
// the service's config, and returns a channel to a timer for when it should be
// restarted. The "timer" returned will have been started.
func (s *Service) scheduleRestart(status api.ServiceStatus) (timer <-chan time.Time) {
	var restart bool

	switch s.config.Type {
	case config.TypeOneshot:
		restart = s.oneshotShouldRestart(status)
	case config.TypeSimple:
		restart = s.simpleShouldRestart(status)
	default:
		logger.Warnf(s.Name(), "unable to determine restart behavior, unknown service type: %s", s.config.Type)
	}

	if restart {
		logger.Debugln(s.Name(), fmt.Sprintf("restarting in %d second(s)", uint64(s.config.RestartDur.Seconds())))
		timer = time.After(s.config.RestartDur)
		return
	}
	return
}

// Run the ExecOnFailure cmd and close the given channel when done
func (s *Service) execOnFail(done chan struct{}) {
	s.runCmd(s.config.ExecOnFailureCmd, s.config.ExecOnFailureCmdArgs)
	close(done)
}

==> superd-0.3.2.tar.gz: superd-0.3.2/internal/service/service.go <==
// Copyright 2022 Clayton Craft <clayton@craftyguy.net>
// SPDX-License-Identifier: GPL-3.0-or-later

package service

import (
	"fmt"
	"os"
	"sync"
	"time"

	goCmd "github.com/go-cmd/cmd"
	"sr.ht/~craftyguy/superd/internal/cmd"
	"sr.ht/~craftyguy/superd/internal/config"
	"sr.ht/~craftyguy/superd/internal/env"
	"sr.ht/~craftyguy/superd/internal/log"
	"sr.ht/~craftyguy/superd/pkg/api"
)

type Service struct {
	Start   chan bool
	Stop    chan bool
	Details chan api.ServiceDetails
	Config  chan *config.ServiceConfig

	// func for one-shot getting command status
	cmdStatus func() goCmd.Status

	before struct {
		list []string
		sync.RWMutex
	}
	after struct {
		list []string
		sync.RWMutex
	}

	name struct {
		value string
		sync.RWMutex
	}

	autostart struct {
		value bool
		sync.RWMutex
	}

	config *config.ServiceConfig
}

// whether or not the service should be autostarted
func (s *Service) Autostart() bool {
	s.autostart.RLock()
	defer s.autostart.RUnlock()
	return s.autostart.value
}

func (s *Service) Name() string {
	s.name.RLock()
	defer s.name.RUnlock()
	return s.name.value
}

func (s *Service) AddAfter(name string) {
	s.after.Lock()
	defer s.after.Unlock()
	for _, n := range s.after.list {
		if n == name {
			return
		}
	}
	s.after.list = append(s.after.list, name)
}

func (s *Service) AddBefore(name string) {
	s.before.Lock()
	defer s.before.Unlock()
	for _, n := range s.before.list {
		if n == name {
			return
		}
	}
	s.before.list = append(s.before.list, name)
}

func (s *Service) After() []string {
	s.after.RLock()
	defer s.after.RUnlock()
	// return a copy
	return append([]string{}, s.after.list...)
}

func (s *Service) Before() []string {
	s.before.RLock()
	defer s.before.RUnlock()
	// return a copy
	return append([]string{}, s.before.list...)
}

var logger *log.Logger

func Init(name string) (s *Service) {
	logger = log.Get()

	s = &Service{
		Start:   make(chan bool),
		Stop:    make(chan bool),
		Details: make(chan api.ServiceDetails),
		Config:  make(chan *config.ServiceConfig),
	}
	s.name.Lock()
	defer s.name.Unlock()
	s.name.value = name

	return
}

func (s *Service) runCmd(c cmd.Commander) (err error) {

	var e []string
	// set service-specific env. from config
	for k, v := range s.config.Env {
		e = append(e, fmt.Sprintf("%s=%s", k, v))
	}
	// add "global" environment
	for v := range env.Init().IterVars() {
		e = append(e, v)
	}
	// import all variables from current env
	e = append(e, os.Environ()...)

	return c.Start(e)
}

// make cmd.NewCommand overridable in testing
var newCommand = cmd.NewCommand

// startup service routine, e.g. running ExecStartPre, ExecStart. Returns
// channels for receiving the command object if successful or service status on
// failure.
func (s *Service) startup() (cmdChan chan cmd.Commander, startStatus chan api.ServiceStatus) {
	startStatus = make(chan api.ServiceStatus)
	cmdChan = make(chan cmd.Commander)

	tries := 0
	maxTries := s.config.StartLimitBurst

	go func() {
		defer close(startStatus)
		defer close(cmdChan)

		// ExecStartPre
		if s.config.ExecStartPreCmd != "" {
			preCmd := newCommand(s.config.ExecStartPreCmd, s.config.ExecStartPreCmdArgs)
			if err := s.runCmd(preCmd); err != nil {
				startStatus <- api.ServiceFailed
				return
			}
		}

		var startCmd cmd.Commander
		// ExecStart
	start_loop:
		for {
			if tries >= maxTries {
				logger.Errorf(s.Name(), "failed to start service: exceeded the max (%d) number of tries", maxTries)
				startStatus <- api.ServiceFailed
				return
			}

			startCmd = newCommand(s.config.Cmd, s.config.CmdArgs)
			if err := s.runCmd(startCmd); err != nil {
				switch err.(type) {
				case cmd.ErrCmdExitedNonZero:
				case cmd.ErrUnhandledStatus:
				case cmd.ErrCmdExited:
					// oneshot is expected to exit
					if s.config.Type == config.TypeOneshot {
						break start_loop
					}
				case cmd.ErrTimeout:
				default:
					logger.Errorln(s.Name(), "unknown failure when trying to start service: "+err.Error())
				}
				logger.Errorln(s.Name(), "failed to start service: "+err.Error())
			} else {
				break
			}

			tries++
			logger.Errorf(s.Name(), "try %d of %d\n", tries, maxTries)
			time.Sleep(s.config.StartLimitDur)
		}

		// ExecStartPost
		if s.config.ExecStartPostCmd != "" {
			postCmd := newCommand(s.config.ExecStartPostCmd, s.config.ExecStartPostCmdArgs)
			if err := s.runCmd(postCmd); err != nil {
				startStatus <- api.ServiceFailed
				return
			}
		}

		// send the cmd from ExecStart
		cmdChan <- startCmd
	}()

	return
}

func (s *Service) Uptime() int64 {
	if s.cmdStatus == nil {
		return 0
	}
	return int64(s.cmdStatus().Runtime)
}

func (s *Service) Run(quit <-chan struct{}) {

	var (
		startStatus    <-chan api.ServiceStatus
		onFailComplete chan struct{}
		restartTimer   <-chan time.Time
		details        api.ServiceDetails

		// used for receiving the started service process details
		cmdChan <-chan cmd.Commander
		command cmd.Commander

		// command streams
		cmdStatus <-chan goCmd.Status
		cmdStdout <-chan string
		cmdStderr <-chan string
	)
	details.Status = api.ServiceStopped

	for {
		select {

		// Start (process) requested
		case <-s.Start:
			if details.Status != api.ServiceStarted && details.Status != api.ServiceStarting {
				cmdChan, startStatus = s.startup()
				details.Status = api.ServiceStarting
			}

		// new service process info received (e.g. from startup())
		case command = <-cmdChan:
			startStatus = nil
			cmdChan = nil
			streams := command.Streams()
			cmdStatus = streams.Status
			cmdStdout = streams.Stdout
			cmdStderr = streams.Stderr
			details.Status = api.ServiceStarted
			details.Pid = command.Status().PID
			s.cmdStatus = command.Status

		// service startup status
		case ss := <-startStatus:
			startStatus = nil
			cmdStatus = nil
			cmdChan = nil
			switch ss {
			case api.ServiceFailed:
				details.Status = api.ServiceFailed
				// Run execOnFailure
				if s.config.ExecOnFailureCmd != "" {
					// Note: a restart is not scheduled until after the
					// ExecOnFailure command has completed
					onFailComplete = make(chan struct{})
					s.execOnFail(onFailComplete)
				} else {
					restartTimer = s.scheduleRestart(details.Status)
				}
			}

		// Stop (process) requested
		case <-s.Stop:
			// Cancel restart and start timer in so that it doesn't trigger and
			// restart the service
			restartTimer = nil

			// set channel to nil to avoid triggering a restart
			cmdStatus = nil
			details.Status = api.ServiceStopped
			s.cmdStatus = nil

			// guard against many rapid calls to stop
			if command != nil {
				if err := command.Stop(); err != nil {
					logger.Errorln(s.Name(), "stop() returned error: "+err.Error())
					details.Status = api.ServiceFailed
				}
				command = nil

			}

		// Quit (goroutine) requested
		case <-quit:
			// Cancel restart and start timer in so that it doesn't trigger and
			// restart the service
			restartTimer = nil

			// set channel to nil to avoid triggering a restart
			cmdStatus = nil
			if command != nil {
				if err := command.Stop(); err != nil {
					logger.Errorln(s.Name(), "failed to stop: "+err.Error())
				}
			}
			return

		// Send details/status
		case s.Details <- details:

		// Status from running process
		case status := <-cmdStatus:
			// status from go-cmd
			if status.Complete {
				logger.Debugln(s.Name(), "exited with: "+fmt.Sprint(status.Exit))
				s.cmdStatus = nil

				if status.Exit != 0 {
					details.Status = api.ServiceFailed

					// fire off any execOnFailure
					if s.config.ExecOnFailureCmd != "" {
						onFailComplete = make(chan struct{})
						s.execOnFail(onFailComplete)

						// don't schedule a restart until after the
						// ExecOnFailure command has completed
						continue
					}
				} else if s.config.Type == config.TypeOneshot {
					// show 'started' for oneshot services
					details.Status = api.ServiceStarted
				} else {
					details.Status = api.ServiceStopped
				}

				// Restart scheduled if necessary
				restartTimer = s.scheduleRestart(details.Status)
			} else {
				logger.Warnln(s.Name(), "received unhandled status: "+fmt.Sprint(s))
			}

		// onFailStatus -- from ExecOnFailure
		case <-onFailComplete:
			onFailComplete = nil
			// Restart scheduled if necessary
			restartTimer = s.scheduleRestart(details.Status)

		// Restart timer
		case <-restartTimer:
			go func() {
				// TODO: ewwwww
				s.Start <- true
			}()

		// Log output from command
		case line, open := <-cmdStderr:
			if !open {
				cmdStderr = nil
			} else {
				logger.Debugln(s.Name(), line)
			}
		case line, open := <-cmdStdout:
			if !open {
				cmdStdout = nil
			} else {
				logger.Debugln(s.Name(), line)
			}

		// New config received
		case c := <-s.Config:
			s.config = c

			// update details. note that this is only updating fields that
			// aren't changed by starting/stopping the service, or other
			// runtime events related to the supervised service.
			details.Autostart = c.Autostart
			details.ConfigPath = c.Path
			details.Type = string(c.Type)
			details.Description = c.Description

			// update the autostart field, which is thread safe
			s.autostart.Lock()
			s.autostart.value = c.Autostart
			s.autostart.Unlock()
		}
	}
}

func (s *Service) oneshotShouldRestart(status api.ServiceStatus) bool {
	// oneshot should only restart if it is set to Always or OnFailure, and
	// only when the status is "failed"
	if status == api.ServiceFailed {
		if s.config.Restart == config.RestartAlways || s.config.Restart == config.RestartFailure {
			return true
		}
	}
	return false
}

func (s *Service) simpleShouldRestart(status api.ServiceStatus) bool {
	if status == api.ServiceFailed {
		if s.config.Restart == config.RestartAlways || s.config.Restart == config.RestartFailure {
			return true
		}
	} else if status == api.ServiceStopped {
		if s.config.Restart == config.RestartAlways {
			return true
		}
	}

	return false
}

// Checks whether a restart should be scheduled based on the given status and
// the service's config, and returns a channel to a timer for when it should be
// restarted. The "timer" returned will have been started.
func (s *Service) scheduleRestart(status api.ServiceStatus) (timer <-chan time.Time) {
	var restart bool

	switch s.config.Type {
	case config.TypeOneshot:
		restart = s.oneshotShouldRestart(status)
	case config.TypeSimple:
		restart = s.simpleShouldRestart(status)
	default:
		logger.Warnf(s.Name(), "unable to determine restart behavior, unknown service type: %s", s.config.Type)
	}

	if restart {
		logger.Debugln(s.Name(), fmt.Sprintf("restarting in %d second(s)", uint64(s.config.RestartDur.Seconds())))
		timer = time.After(s.config.RestartDur)
		return
	}
	return
}

// Run the ExecOnFailure cmd and close the given channel when done
func (s *Service) execOnFail(done chan struct{}) {
	c := newCommand(s.config.ExecOnFailureCmd, s.config.ExecOnFailureCmdArgs)
	if err := s.runCmd(c); err != nil {
		logger.Errorf(s.Name(), "failed to run ExecOnFailure %q: %s\n", s.config.ExecOnFailureCmd, err)
	}
	close(done)
}
